"""Module for any LR-style experiment."""
import pandas as pd
from sklearn.model_selection import GridSearchCV
from sklearn.linear_model import LogisticRegression

from enso.experiment.grid_search import GridSearch


class LogisticRegressionCV(GridSearch):
    """Implementation of a grid-search optimized Logistic Regression model."""

    def __init__(self, *args, **kwargs):
        """Initialize internal classifier."""
        super().__init__(*args, **kwargs)
        self.base_model = LogisticRegression
        self.param_grid = {
            'penalty': ['l2'],
            'C': [0.1, 1.0, 10., 100., 1000.],
        }



"""
From: https://www.figure-eight.com/data-for-everyone/

In a variation on the popular task of sentiment analysis, this dataset contains labels for the emotional content (such as happiness, sadness, and anger) of texts. Hundreds to thousands of examples across 13 labels.
"""

from enso.download import generic_download

if __name__ == "__main__":
    generic_download(
        url="https://www.figure-eight.com/wp-content/uploads/2016/07/text_emotion.csv",
        text_column="content",
        target_column="sentiment",
        filename="Emotion.csv"
    )



"""
Base class definition for all training data samples.
Samplers aid transfer learning methods by attempting to request
labeled examples that are more informative than a simple random sample of training data
"""

import numpy as np
import random
from sklearn.metrics.pairwise import pairwise_distances
from enso.config import TEST_SETUP
from enso.utils import get_plugins


def sample(sampler, data, train_labels, train_indices, train_size):
    sampler = Sampler._class_for(sampler)(data, train_labels, train_indices, train_size)
    return sampler.sample()


class Sampler(object):
    """
    Base class for all `Sampler`'s

    :param data: pd.Series of feature vectors
    :param train_labels: pd.Series of targets
    :param train_indices: pd.Series of example indices
    :param train_size: int number of examples to select
    """

    def __init__(self, data, train_labels, train_indices, train_size):
        """
        Given a set of training data and a desired size, uses a strategy to
        select a subset of datapoints to use as labeled training data.
        All arguments are taken at initialization.  The :func:`sample` function
        may be called multiple times on a single `Sampler` object.


        """
        self.data = data
        self.train_labels = train_labels
        self.train_indices = train_indices
        self.train_size = train_size
        if len(self.classes) > train_size:
            raise ValueError("The train size can not be smaller than the number of classes.")

    @property
    def classes(self):
        if not hasattr(self, '_classes'):
            self._classes = set(self.train_labels)
        return self._classes

    def _choose_starting_points(self):
        """
        Ensures a minimum of one label per class is chosen
        """
        points = []
        for cls in self.classes:
            indices = [i for i, val in enumerate(self.train_labels) if val == cls]
            index = random.choice(indices)
            points.append(self.train_indices[index])
        return points

    @property
    def distances(self):
        if not hasattr(self, '_distances'):
            self._distances = pairwise_distances(self.points, metric=self.DISTANCE_FUNCTION)
        return self._distances

    @property
    def points(self):
        return [np.array(point) for point in self.data]

    def sample(self):
        """
        Given the settings provided at initialization, apply the provided sampling strategy

        :returns: np.array of example indices selected by `Sampler`.
        """
        raise NotImplementedError

    @classmethod
    def _class_for(cls, sampler_string):
        return get_plugins("sample", set([sampler_string]), return_class=True)[0]


import spacy
import numpy as np

from spacy.cli.download import download
from enso.featurize import Featurizer



class SpacyCNNFeaturizer(Featurizer):
    """
    Featurizer that embeds documents using the mean of spacy's text CNN embedding model.

    """
    def load(self):
        """
        If the pre-trained `en_core_web_lg` model is not already stored on disk, it will be automatically downloaded
        as part of :func:`load()`. Note that the download process may require sudo permissions depending on your python package settings.
        """
        try:
            self.nlp = spacy.load('en_core_web_lg')
        except OSError:
            download('en_core_web_lg')
            self.nlp = spacy.load('en_core_web_lg')

    def featurize(self, x):
        return self.nlp(x).vector

    def featurize_batch(self, X, **kwargs):
        return [self.nlp(x).vector for x in X]


class SpacyGloveFeaturizer(Featurizer):
    """
    Featurizer that embeds documents using the mean of a document's glove vectors.
    """
    def load(self):
        """
        If the pre-trained `en_vectors_web_lg` model is not already stored on disk, it will be automatically downloaded
        as part of :func:`load()`. Note that the download process may require sudo permissions depending on your python package settings.
        """
        try:
            self.nlp = spacy.load('en_vectors_web_lg')
        except OSError:
            download('en_vectors_web_lg')
            self.nlp = spacy.load('en_vectors_web_lg')

    def featurize(self, x):
        return self.nlp(x).vector

    def featurize_batch(self, X, **kwargs):
        return [self.nlp(x).vector for x in X]



"""
Entrypoint for featurizing datasets according to the specifications of `config.py`.
"""

import logging

import pandas as pd

import concurrent.futures
from concurrent.futures import ThreadPoolExecutor, ProcessPoolExecutor
from enso.config import FEATURIZERS, DATA, N_CORES
from enso.utils import get_plugins, feature_set_location, BaseObject
from sklearn.externals import joblib


class Featurization(object):
    """
    Orchestrates the application of the selected featurizers to a set of featurizers
    according to the settings specified in `config.py`
    """

    def __init__(self):
        """
        Responsible for searching featurizer module and importing those specified in config.
        """
        self.featurizers = get_plugins('featurize', FEATURIZERS)

    def _run(self, POOL):
        futures = {}
        for featurizer in self.featurizers:
            featurizer.load()
            for dataset_name in DATA:
                dataset = self._load_dataset(dataset_name)
                logging.info("Featurizing {} with {}....".format(dataset_name, featurizer.__class__.__name__))
                future = POOL.submit(featurizer.generate, dataset, dataset_name)
                futures[future] = (featurizer, dataset_name)

        for future in concurrent.futures.as_completed(futures):
            featurizer, dataset_name = futures[future]
            try:
                future.result()
                logging.info("Completed featurization of dataset `{dataset_name}` with featurizer `{featurizer}`.".format(
                    dataset_name=dataset_name,
                    featurizer=featurizer.__class__.__name__
                ))
            except Exception as e:
                logging.exception("Failed featurization of dataset `{dataset_name}` with featurizer `{featurizer}`.".format(
                    dataset_name=dataset_name,
                    featurizer=featurizer.__class__.__name__
                ))

    def run(self, n_jobs=N_CORES):
        """
        Responsible for ensuring every active featurizer is applied to every active dataset.
        Featurization is parallelized by default, with `n_jobs` set to number of cores specified in
        config.py.  If n_jobs is set to 1 or less, featurization is run in a single threaded setting.
        """
        if n_jobs > 1:
            POOL = ProcessPoolExecutor(n_jobs)
            self._run(POOL)
        else:
            self._run(ThreadPoolExecutor(1))

    @staticmethod
    def _load_dataset(dataset_name):
        """Responsible for finding datasets and reading them into dataframes."""
        df = pd.read_csv("Data/%s.csv" % dataset_name)
        if 'Text' not in df:
            raise ValueError("File: %s has no column 'Text'" % dataset_name)
        if 'Target' not in df:
            raise ValueError("File %s has no column 'Target'" % dataset_name)
        return df


class Featurizer(BaseObject):
    """
    Base class for building featurizers.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def load(self):
        """
        Method called in flow of `python -m enso.featurize` to prevent loading
        pre-trained models into memory on file import.

        If loading a pre-trained model into memory is not required, `Featurizer.load()`
        defaults to `pass`.
        """
        pass

    def generate(self, dataset, dataset_name):
        """
        Given a `dataset` pandas DataFrame and string `dataset_name`,
        add column `"Features"` to the provided `pd.DataFrame` and serialize the result
        to the results folder listed in `config.py`.

        If a given featurizer exposes a :func:`featurize_batch` method, that method will be
        called to perform featurization.  Otherwise, :class:`Featurizer`'s will fall
        back to calling :func:`featurize` on each individual example.

        :param dataset: `pd.DataFrame` object that must contain a `Text` column.
        :param dataset_name: `str` name to use as a save location in the `config.FEATURES_DIRECTORY`.
        """
        features = []
        try:
            features = self.featurize_batch(dataset['Text'])
        except (NotImplementedError, AttributeError):
            try:
                features = [self.featurize(entry) for entry in dataset['Text']]
            except (NotImplementedError, AttributeError):
                raise NotImplementedError("""
                    Featurizers must implement the featurize_list, or the featurize method
                """)
        new_dataset = dataset.copy()  # Don't want to modify the underlying dataframe
        new_dataset['Features'] = features
        self._write(new_dataset, dataset_name)

    def _write(self, featurized_dataset, dataset_name):
        """Responsible for taking a featurized dataset and writing it out to the filesystem."""
        dump_location = feature_set_location(dataset_name, self.__class__.__name__)
        joblib.dump(featurized_dataset, dump_location)

    def featurize_batch(self, X, batch_size=32):
        """
        :param X: `pd.Series` that contains raw text to featurize
        :param batch_size: int number of examples to process per batch
        :returns: list of np.ndarray representations of text
        """
        raise NotImplementedError

    def featurize(self, text):
        """
        :param text: text of a singular example
        :returns: `np.ndarray` representation of text
        """
        raise NotImplementedError



from . import Sampler
import numpy as np


class Random(Sampler):
    """
    Randomly selects examples from the training dataset.

    :param data: pd.Series of feature vectors
    :param train_labels: pd.Series of targets
    :param train_indices: pd.Series of example indices
    :param train_size: int number of examples to select
    """

    def sample(self):
        """
        Randomly samples feature vectors.

        :returns: np.array of example indices selected by random sampling
        """
        points = self._choose_starting_points()
        return points + list(np.random.choice(self.train_indices, self.train_size - len(points), replace=False))



import logging

from enso.config import EXPERIMENT_NAME

from . import Experimentation

if __name__ == "__main__":
    logging.basicConfig(level=logging.INFO)
    logging.info("Experimentation Started...")
    experimentation = Experimentation(EXPERIMENT_NAME)
    logging.info("Running Experiments...")
    experimentation.run_experiments()



"""General util functions."""
import os
import os.path
import inspect
from time import strptime
from importlib import import_module

import pandas as pd

from enso.config import RESULTS_DIRECTORY, FEATURES_DIRECTORY


def feature_set_location(dataset_name, featurizer_name):
    """Responsible for generating filenames for generated feature sets."""
    base_dir, _, filename = dataset_name.rpartition('/')
    write_location = "%s/%s/" % (FEATURES_DIRECTORY, base_dir)
    dump_name = "%s_%s_features.csv" % (filename, featurizer_name)
    return write_location + dump_name


def get_plugins(plugin_dir, match_names, return_class=False):
    """Responsible for grabbing objects from specified plugin directories."""
    full_plugin_dir_path = os.path.abspath(os.path.join(os.path.dirname(__file__), plugin_dir))
    root, dirs, files = next(os.walk(full_plugin_dir_path))

    relevant_classes = []
    for filename in files:
        if not filename.endswith('.py') or '__' in filename:
            continue
        module_name = filename.rpartition('.')[0]
        import_path = "%s.%s.%s" % ("enso", plugin_dir, module_name)
        mod = import_module(import_path)
        for name, obj in inspect.getmembers(mod):
            if inspect.isclass(obj) and obj.__name__ in match_names:
                relevant_classes.append(obj)

    names = set(item.__name__ for item in relevant_classes)
    if names != match_names:
        raise ValueError("""
            Config doesn't match classes present.\n%s: %s\nConfig: %s
        """ % (plugin_dir, names, match_names))

    if return_class:
        return relevant_classes
    else:
        return [relevant_class() for relevant_class in relevant_classes]


def get_all_experiment_runs():
    """Grab all experiment runs and return a list sorted by date."""
    dirs = [item for item in os.listdir(RESULTS_DIRECTORY)]
    timed_dirs = []
    for directory in dirs:
        try:
            dt = strptime(directory, "%Y-%m-%d_%H-%M-%S")
            timed_dirs.append((directory, dt))
        except ValueError:
            pass
    timed_dirs.sort(key=lambda pair: pair[1], reverse=True)
    dirs = [pair[0] for pair in timed_dirs]
    return dirs


def labels_to_binary(target_list):
    """
    Convert a list of labels into a pandas dataframe appropriate for metric calculations.

    Example: labels_to_binary('apple', ['apple', 'orange']) ->
    pd.DataFrame({'apple': [1, 0], 'orange': [0, 1]})
    """
    full_mapping = {}
    for target in set(target_list):
        full_mapping[target] = [int(item == target) for item in target_list]
    return pd.DataFrame(full_mapping)


class BaseObject(object):
    """Base object for all plugins."""

    def name(self):
        """
        Prints the name of the current class to aid logging and result formatting.
        """
        return self.__class__.__name__



"""
From: https://www.figure-eight.com/data-for-everyone/

Contributors read sentences in which both a chemical (like Aspirin) and a disease (or side-effect) were present. They then determined if the chemical directly contributed to the disease or caused it. Dataset includes chemical names, disease name, and aggregated judgments of five (as opposed to the usual three) contributors.
"""

from enso.download import generic_download

if __name__ == "__main__":
    generic_download(
        url="https://www.figure-eight.com/wp-content/uploads/2016/03/chemicals-and-disease-DFE.csv",
        text_column="form_sentence",
        target_column="verify_relationship",
        filename="ChemicalDiseaseCauses.csv"
    )



import logging

from . import Featurization

if __name__ == "__main__":
    logging.basicConfig(level=logging.INFO)
    logging.info("Initializing Featurizers...")
    featurization = Featurization()
    logging.info("Converting Datasets...")
    featurization.run()



"""
From: https://www.figure-eight.com/data-for-everyone/

Contributors viewed tweets regarding a variety of left-leaning issues like legalization of abortion, feminism, Hillary Clinton, etc. They then classified if the tweets in question were for, against, or neutral on the issue (with an option for none of the above). After this, they further classified each statement as to whether they expressed a subjective opinion or gave facts.
"""
from enso.download import generic_download


if __name__ == "__main__":
    generic_download(
        url="https://www.figure-eight.com/wp-content/uploads/2016/03/progressive-tweet-sentiment.csv",
        text_column="tweet",
        target_column="q1_from_reading_the_tweet_which_of_the_options_below_is_most_likely_to_be_true_about_the_stance_or_outlook_of_the_tweeter_towards_the_target",
        filename="PoliticalTweetAlignment.csv"
    )

    generic_download(
        url="https://www.figure-eight.com/wp-content/uploads/2016/03/progressive-tweet-sentiment.csv",
        text_column="tweet",
        target_column="q2_which_of_the_options_below_is_true_about_the_opinion_in_the_tweet",
        filename="PoliticalTweetSubjectivity.csv"
    )

    generic_download(
        url="https://www.figure-eight.com/wp-content/uploads/2016/03/progressive-tweet-sentiment.csv",
        text_column="tweet",
        target_column="target",
        filename="PoliticalTweetTarget.csv"
    )



from collections import Counter, defaultdict
import random

import numpy as np


def oversample(X, y, max_ratio=50):
    """
    Ensure each class occurs with approximately even frequency in the training set by
    duplicating examples from relatively rare classes.

    :param X: `np.ndarray` of input features
    :param y: `np.ndarray` of corresponding targets
    :param max_ratio: input examples should be duplicated no more than this amount
    """
    X, y = np.asarray(X), np.asarray(y)
    class_counts = Counter(y)
    max_count = max(class_counts.values())
    desired_counts = {
        class_name: min(
            max_count, max_ratio * class_count)
        for class_name, class_count in class_counts.items()
    }

    idxs_by_y = defaultdict(list)
    for i, element in enumerate(y):
        idxs_by_y[element].append(i)

    idx_sample = []
    for class_name in class_counts:
        sample_idxs = np.random.choice(idxs_by_y[class_name], desired_counts[class_name], replace=True).tolist()
        idx_sample.extend(sample_idxs)

    random.shuffle(idx_sample)

    return X[idx_sample].tolist(), y[idx_sample].tolist()


def resample(resample_type, train_data, train_labels):
    if resample_type.lower() == 'none':
        return train_data, train_labels
    elif resample_type == 'RandomOverSampler':
        train_data, train_labels = oversample(train_data, train_labels)
        return train_data, train_labels
    else:
        raise Exception("Invalid resample_type: {}".format(resample_type))



from abc import abstractmethod

from sklearn.model_selection import GridSearchCV
import pandas as pd

from enso.experiment import ClassificationExperiment


class GridSearch(ClassificationExperiment):
    """
    Base class for classification models that select hyperparameters via cross validation.
    Assumes the `base_model` property set on child classes inherits from
    `sklearn.base.BaseEstimator` and implements `predict_proba` and `score`.

    :cvar base_model: Class name of base model, must be set by child classes.
    :cvar param_grid: Dictionary that maps from class paramaters to an array of values to search over.  Must be set by child classes.
    """

    param_grid = {}
    base_model = None

    def __init__(self, *args, **kwargs):
        """Initialize internal classifier."""
        super().__init__(*args, **kwargs)
        self.best_model = None

    def fit(self, X, y):
        """
        Runs grid search over `self.param_grid` on `self.base_model` to optimize hyper-parameters using
        KFolds cross-validation, then retrains using the selected parameters on the full training set.

        :param X: `np.ndarray` of input features sampled from training data.
        :param y: `np.ndarray` of corresponding targets sampled from training data.
        """
        classifier = GridSearchCV(
            self.base_model(),
            param_grid=self.param_grid
        )
        classifier.fit(X, y)

        self.best_model = self.base_model(**classifier.best_params_)
        self.best_model.fit(X, y)

    def predict(self, X, **kwargs):
        """Predict results on test set based on current internal model."""
        labels = self.best_model.classes_
        probabilities = self.best_model.predict_proba(X)
        return pd.DataFrame({label: probabilities[:, i] for i, label in enumerate(labels)})



import logging

from enso.config import EXPERIMENT_NAME
from enso.visualize import Visualization

if __name__ == "__main__":
    logging.info('Loading Results...')
    visualization = Visualization(EXPERIMENT_NAME)
    logging.info('Painting Pretty Pictures...')
    visualization.visualize()



"""Module for any NB-style experiment."""
import pandas as pd
from sklearn.model_selection import GridSearchCV
from sklearn.naive_bayes import GaussianNB, MultinomialNB, BernoulliNB

from enso.experiment import Experiment


class NaiveGaussianBayes(Experiment):
    """Basic implementation of a grid-search optimized Logistic Regression."""

    def __init__(self):
        """Initialize internal classifier."""
        self.model = GaussianNB
        self.active_model = None

    def train(self, training_data, training_labels):
        """Run grid search to optimize hyper-parameters, then trains the final model."""
        self.active_model = self.model()
        self.active_model.fit(training_data, training_labels)

    def predict(self, dataset):
        """Predict results on test set based on current internal model."""
        labels = self.active_model.classes_
        probabilities = self.active_model.predict_proba(dataset)
        return pd.DataFrame({label: probabilities[:, i] for i, label in enumerate(labels)})

class NaiveMultinomialBayes(Experiment):
    """Basic implementation of a grid-search optimized Logistic Regression."""

    def __init__(self):
        """Initialize internal classifier."""
        self.model = MultinomialNB
        self.active_model = None
        self.param_grid = {'alpha': [0, 0.25, 0.5, 0.75, 1], 'fit_prior': [True, False]}

    def train(self, training_data, training_labels):
        """Run grid search to optimize hyper-parameters, then trains the final model."""
        classifier = GridSearchCV(self.model(), self.param_grid)
        classifier.fit(training_data, training_labels)
        # Train with model with ideal params on whole training set
        self.active_model = self.model(**classifier.best_params_)
        self.active_model.fit(training_data, training_labels)

    def predict(self, dataset):
        """Predict results on test set based on current internal model."""
        labels = self.active_model.classes_
        probabilities = self.active_model.predict_proba(dataset)
        return pd.DataFrame({label: probabilities[:, i] for i, label in enumerate(labels)})

class NativeBernoulliBayes(Experiment):
    """Basic implementation of a grid-search optimized Logistic Regression."""

    def __init__(self):
        """Initialize internal classifier."""
        self.model = BernoulliNB
        self.active_model = None
        self.param_grid = {'alpha': [0, 0.25, 0.5, 0.75, 1], 'fit_prior': [True, False], 'binarize': [True, False]}

    def train(self, training_data, training_labels):
        """Run grid search to optimize hyper-parameters, then trains the final model."""
        classifier = GridSearchCV(self.model(), self.param_grid)
        classifier.fit(training_data, training_labels)
        # Train with model with ideal params on whole training set
        self.active_model = self.model(**classifier.best_params_)
        self.active_model.fit(training_data, training_labels)

    def predict(self, dataset):
        """Predict results on test set based on current internal model."""
        labels = self.active_model.classes_
        probabilities = self.active_model.predict_proba(dataset)
        return pd.DataFrame({label: probabilities[:, i] for i, label in enumerate(labels)})


"""Module for all metrics that one might want for evaluation."""
import abc

from enso.utils import BaseObject


class Metric(BaseObject):
    """Base class for all Metrics."""

    @abc.abstractmethod
    def evaluate(self, ground_truth, result):
        """
        General metric endpoint for generating results.

        This should be implemented by every metric class. No exceptions.
        """
        raise NotImplementedError


class ClassificationMetric(Metric):
    """Base class for classification metrics."""

    pass


class RegressionMetric(Metric):
    """Base class for regression metrics."""

    pass


class MatchingMetric(Metric):
    """Base class for matching metrics."""

    pass



"""Module for creating facet visualization."""
import os.path
import json

import seaborn as sns
import matplotlib.patches as patches
from matplotlib import pyplot as plt

from enso.visualize import ClassificationVisualizer
from enso.config import RESULTS_DIRECTORY


class FacetGridVisualizer(ClassificationVisualizer):
    """
    Create a grid of line graphs based on the value of `config.VISUALIZATION_OPTIONS`
    """

    def visualize(
        self,
        results,
        x_tile,
        y_tile,
        x_axis,
        y_axis,
        lines,
        results_id=None,
        filename='FacetGridVisualizer',
        **kwargs
    ):
        """
        Create a tiled visualization of experiment results.

        :param results: pd.DataFrame of results, loaded from results .csv file.
        :param x_tile: string name of DataFrame column to vary over the x axis of the grid of line graphs
        :param y_tile: string name of DataFrame column to vary over the y axis of the grid of line graphs
        :param x_axis: string name of DataFrame column to plot on the x axis within each individual line graph
        :param y_axis: string name of DataFrame column to plot on the y axis within each individual line graph
        :param lines: string name or list of DataFrame column string names displayed as separate lines within each graph.
                      Providing multiple values means that each unique combination of values will be displayed as a single line.
        :param results_id: string name of folder to save resulting visual in, relative to the root of the results directory
        :param filename: filename (excluding filetype) to use when saving visualization.  Value is relative to folder specified by results_id.
        """
        sns.set(style="ticks", color_codes=True)

        if isinstance(lines, (tuple, list)):
            results['key'] = results[lines].apply(lambda x: ','.join(x), axis=1)
            lines = 'key'

        n_lines = len(results[lines].unique())
        rc = {'lines.linewidth': 1, 'lines.markersize': 1}
        sns.set_context("paper", rc=rc)
        grid = sns.FacetGrid(
            results,
            col=x_tile,
            row=y_tile,
            hue=lines,
            size=5,
            legend_out=False,
            margin_titles=True,
            sharey=False,
            sharex=False,
            palette=sns.color_palette("hls", n_lines)
        )
        grid = grid.map(sns.pointplot, x_axis, y_axis)
        colors = sns.color_palette("hls", n_lines).as_hex()[:len(grid.hue_names)]
        handles = [patches.Patch(color=col, label=label) for col, label in zip(colors, grid.hue_names)]
        plt.legend(handles=handles)
        plt.tight_layout()
        filename = os.path.join(RESULTS_DIRECTORY, results_id, "{}.png".format(filename))
        plt.savefig(filename)



from . import Sampler
import numpy as np


class KCenter(Sampler):
    """
    Randomly selects an example from each class to use as "centers",
    then selects points that are maximally distant from any given center
    to use as training examples.

    :param data: pd.Series of feature vectors
    :param train_labels: pd.Series of targets
    :param train_indices: pd.Series of example indices
    :param train_size: int number of examples to select
    """
    DISTANCE_FUNCTION = "euclidean"

    def sample(self):
        """
        Applies the KCenter sampling strategy.

        :returns: np.array of example indices selected by `KCenter`.
        """
        centers = self._choose_starting_points()
        while len(centers) < self.train_size:
            mins = np.min(self.distances[centers], axis=0)
            center = np.argmax(mins)
            centers.append(center)
        return centers


"""
From: https://www.figure-eight.com/data-for-everyone/

Contributors looked at a single sentence and rated its emotional content based on Plutchik’s wheel of emotions. 18 emotional choices were presented to contributors for grading.
"""

from enso.download import generic_download

if __name__ == "__main__":
    generic_download(
        url="https://www.figure-eight.com/wp-content/uploads/2016/03/primary-plutchik-wheel-DFE.csv",
        text_column="sentence",
        target_column="emotion",
        filename="DetailedEmotion.csv"
    )



"""File for storing the featurizers that indico offers via API."""
from indicoio.custom import vectorize
import indicoio.config
from tqdm import tqdm

from enso.featurize import Featurizer


class BaseIndicoFeaturizer(Featurizer):
    """
    Base model from which all indico `Featurizer`'s inherit.
    """
    domain = None
    sequence = False

    def featurize_batch(self, X, batch_size=32, **kwargs):
        """
        :param X: `pd.Series` that contains raw text to featurize
        :param batch_size: int number of examples to process per batch
        :returns: list of np.ndarray representations of text
        """
        all_features = []
        for i in tqdm(range(0, len(X), batch_size)):
            chunk_data = list(X[i:i + batch_size])
            all_features.extend(vectorize(
                chunk_data, domain=self.domain, sequence=self.sequence, **kwargs
            ))
        return all_features


class IndicoStandard(BaseIndicoFeaturizer):
    """Featurizer that uses indico's standard features."""
    domain = 'standard'


class IndicoSentiment(BaseIndicoFeaturizer):
    """Featurizer that uses indico's sentiment features."""
    domain = 'sentiment'


class IndicoTopics(BaseIndicoFeaturizer):
    """Featurizer that uses indico's topics features."""
    domain = 'topics'


class IndicoFinance(BaseIndicoFeaturizer):
    """Featurizer that uses indico's finance features."""
    domain = 'finance'


class IndicoTransformer(BaseIndicoFeaturizer):
    """Featurizer that uses indico's transformer features."""
    domain = 'transformer'


class IndicoEmotion(BaseIndicoFeaturizer):
    """Featurizer that uses indico's emotion features."""
    domain = 'emotion'


class IndicoFastText(BaseIndicoFeaturizer):
    """Featurizer that uses indico's fasttext features."""
    domain = 'fasttext'


class IndicoElmo(BaseIndicoFeaturizer):
    """Featurizer that uses indico's finance features."""
    domain = 'elmo'


# NOTE: To remain undocumented for now until further API design decisions are made
class IndicoElmoSequence(BaseIndicoFeaturizer):
    """Featurizer that uses indico's finance features."""
    domain = 'elmo'
    sequence = True


class IndicoTransformerSequence(BaseIndicoFeaturizer):
    """Featurizer that uses indico's transformer sequence features."""
    domain = 'transformer'
    sequence = True


class IndicoStandardSequence(BaseIndicoFeaturizer):
    """Featurizer that uses indico's standard sequence features"""
    domain = 'standard'
    sequence = True



"""
From: https://www.figure-eight.com/data-for-everyone/
Contributors evaluated tweets for belief in the existence of global warming or climate change. The possible answers were “Yes” if the tweet suggests global warming is occurring, “No” if the tweet suggests global warming is not occurring, and “I can’t tell” if the tweet is ambiguous or unrelated to global warming. We also provide a confidence score for the classification of each tweet.
"""

from enso.download import generic_download

def words_to_char(val):
    conversion = {
        "Yes": 'Y',
        "No": 'N'
    }
    converted_val = conversion.get(val, val)
    return converted_val


if __name__ == "__main__":
    generic_download(
        url="https://www.figure-eight.com/wp-content/uploads/2016/03/1377884570_tweet_global_warming.csv",
        text_column="tweet",
        target_column="existence",
        target_transformation=words_to_char,
        filename="GlobalWarming.csv"
    )



"""
Main method for running experiments according to the specifications of `config.py`.
"""
import os
import logging
from shutil import copyfile
from time import gmtime, strftime
from ast import literal_eval
import abc
from functools import wraps
import multiprocessing
import itertools
import json

import concurrent.futures
from concurrent.futures import ProcessPoolExecutor
import pandas as pd
import numpy as np
from sklearn.model_selection import StratifiedShuffleSplit
from sklearn.externals import joblib
from tqdm import tqdm

from enso.resample import resample
from enso.sample import sample
from enso.utils import feature_set_location, get_plugins, BaseObject
from enso.config import FEATURIZERS, DATA, EXPERIMENTS, METRICS, TEST_SETUP, RESULTS_DIRECTORY, N_GPUS, N_CORES

POOL = ProcessPoolExecutor(N_CORES)


class Experimentation(object):
    """Responsible for running experiments configured in config."""

    def __init__(self, name=None):
        """Responsible for gathering and instantiating experiments, featurizers, and metrics."""
        self.name = name
        self.experiments = get_plugins("experiment", EXPERIMENTS, return_class=True)
        self.featurizers = get_plugins("featurize", FEATURIZERS)
        self.metrics = get_plugins("metrics", METRICS)
        self.columns = ['Dataset', 'Featurizer', 'Experiment', 'Metric', 'TrainSize', 'Sampler', 'Resampler', 'Result', 'TrainResult']
        self.results = pd.DataFrame(columns=self.columns)

    def run_experiments(self):
        """Responsible for actually running experiments."""
        futures = {}
        for dataset_name in DATA:
            logging.info("Experimenting on %s dataset" % dataset_name)
            for featurizer in self.featurizers:
                logging.info("Currently using featurizer: %s" % featurizer.name())
                for training_size in TEST_SETUP["train_sizes"]:
                    for sampler in TEST_SETUP["samplers"]:
                        for resampler in TEST_SETUP["resamplers"]:
                            current_setting = {
                                'Dataset': dataset_name,
                                'Featurizer': featurizer.name(),
                                'TrainSize': training_size,
                                'Sampler': sampler,
                                'Resampler': resampler
                            }
                            future = POOL.submit(self._run_experiment, dataset_name, current_setting)
                            futures[future] = current_setting

        for future in concurrent.futures.as_completed(futures):
            current_setting = futures[future]
            try:
                results = future.result()
                logging.info("Finished training for {}".format(current_setting))
            except Exception:
                logging.exception("Exception occurred for {}".format(current_setting))

    def _run_experiment(self, dataset_name, current_setting):
        """Responsible for running all experiments specified in config."""
        results = pd.DataFrame(columns=self.columns)
        dataset = self._load_dataset(dataset_name, current_setting.get("Featurizer"))
        for splitter, target in self._split_dataset(dataset, current_setting['TrainSize']):
            for train_indices, test in splitter:
                train = sample(
                    current_setting['Sampler'],
                    dataset['Features'],
                    list(dataset[target].iloc[train_indices]),
                    train_indices,
                    current_setting['TrainSize']
                )
                for experiment_cls in self.experiments:
                    experiment = experiment_cls()

                    name = experiment.name()
                    internal_setting = {
                        'Experiment': name
                    }
                    internal_setting.update(current_setting)
                    logging.info("Training with settings {}".format(internal_setting))
                    try:
                        # You might find yourself wondering why we're using lists here instead of np arrays
                        # The answer is that pandas sucks.
                        train_set = list(dataset['Features'].iloc[train])
                        train_labels = list(dataset[target].iloc[train])
                        test_set = list(dataset['Features'].iloc[test])
                        test_labels = list(dataset[target].iloc[test])
                        experiment.fit(*resample(current_setting['Resampler'], train_set, train_labels))

                        test_pred = experiment.predict(test_set, subset='TEST')
                        train_pred = experiment.predict(train_set, subset='TRAIN')
                        result = self._measure_experiment(
                            target=test_labels,
                            result=test_pred,
                            train_target=train_labels,
                            train_result=train_pred,
                            internal_setting=internal_setting
                        )
                        self._dump_results(result, experiment_name=self.name)
                    except Exception:
                        logging.exception("Failed to run experiment: {}".format(internal_setting))
        return results

    def _measure_experiment(self, target, result, train_target=None, train_result=None, internal_setting=None, test_key='Result', train_key='TrainResult'):
        """Responsible for recording all metrics specified in config for a given experiment."""
        results = pd.DataFrame(columns=self.columns)
        for metric in self.metrics:
            score = metric.evaluate(target, result)
            full_setting = {"Metric": metric.name(), test_key: score}

            # measure score on train set to help detect overfitting
            if train_target is not None and train_result is not None:
                train_score = metric.evaluate(train_target, train_result)
                full_setting[train_key] = train_score

            full_setting.update(internal_setting)
            full_setting_df = pd.DataFrame.from_records([full_setting])
            results = results.append(full_setting_df, ignore_index=True)
        return results

    def _dump_results(self, results, experiment_name):
        """Responsible for recording config and dumping experiment results in result directory."""
        if not experiment_name:
            experiment_name = strftime("%Y-%m-%d_%H-%M-%S", gmtime())
        result_path = "%s/%s" % (RESULTS_DIRECTORY, experiment_name)
        results = results.dropna()

        if not os.path.exists(result_path):
            os.makedirs(result_path)

        result_file = "%s/Results.csv" % result_path
        header = False if os.path.exists(result_file) else True
        result_fd = open(result_file, 'a')
        results.to_csv(result_fd, header=header, columns=self.columns)

        # The a is for archival, not just a typo
        config_record = "%s/Config.pya" % result_path
        config_path = os.path.abspath(os.path.join(
            os.path.dirname(os.path.dirname(__file__)), 'config.py'
        ))
        copyfile(config_path, config_record)

    @staticmethod
    def _split_dataset(dataset, training_size):
        target_list = [column for column in dataset.columns.values if column.startswith("Target")]
        logging.info("Training with train set of size: %s" % training_size)

        # Sklearn technically offers a train_size parameter that seems like it would be better
        # Unfortunately it doesn't work as expected and locks test size to train size
        test_size = int(len(dataset) * TEST_SETUP["sampling_size"])
        if test_size + training_size > len(dataset):
            raise ValueError("Invalid training size provided.  Training size must be less than {} of dataset size.".format(TEST_SETUP["sampling_size"]))

        splitter = StratifiedShuffleSplit(TEST_SETUP["n_splits"], test_size=test_size)
        for target in target_list:
            # We can use np.zeros because it's stratifying the split
            # based on the categories. Removing the indexing saves time.
            yield splitter.split(np.zeros(len(dataset)), dataset[target]), target

    @staticmethod
    def _load_dataset(dataset_name, featurizer_name):
        """Responsible for loading a given dataset given the dataset_name and featurizer."""
        read_location = feature_set_location(dataset_name, featurizer_name)
        logging.info("Loading Dataset: %s" % read_location)
        return joblib.load(read_location)


class VerifyOutput(BaseObject):
    """Decorator class to add output checks to different experiments classes."""

    def __new__(cls, *args):
        """Wrap predict method with appropriate decorator check."""
        experiment_class = super(VerifyOutput, cls).__new__(cls, *args)
        experiment_class.predict = experiment_class._verify_output(experiment_class.predict)
        return experiment_class


class Experiment(BaseObject):
    """
    Base class for all :class:`Experiment`'s.

    If hyperparameter selection is necessary for a given target model, the :class:`Experiment`
    is responsible for performing hyperparameter selection withing the context of :func:`fit`.

    """

    __metaclass__ = VerifyOutput

    def __init__(self, *args, **kwargs):
        """
        Instantiate a new experiment
        """
        super().__init__(*args, **kwargs)

    @abc.abstractmethod
    def fit(self, X, y):
        """
        Method to begin training of a given target model provided a set of input features
        and corresponding targets.

        :param X: `np.ndarray` of input features sampled from training data.
        :param y: `np.ndarray` of corresponding targets sampled from training data.
        """
        raise NotImplementedError

    @abc.abstractmethod
    def predict(self, X):
        """
        Produce a `pd.DataFrame` object contains target model predictions.

        :param X: `np.ndarray` of input features from test data.
        :return: `pd.DataFrame` of target model predictions

        Prediction format is dependant on class of experiment.
        """
        raise NotImplementedError

    @classmethod
    def _verify_output(cls, function):
        """
        Check predict output to ensure it complies with the experiment type.
        """
        return function


class ClassificationExperiment(Experiment):
    """Base class for classification experiments."""

    @abc.abstractmethod
    def predict(self, X):
        """
        Produce a `pd.DataFrame` object that maps class labels to class probabilities given test inputs.

        :param X: `np.ndarray` of input features from test data.
        :return: `pd.DataFrame` object. Each column should represent a class, and each row should represent an array of probabilities across classes.
        """
        raise NotImplementedError

    @classmethod
    def _verify_output(cls, func):
        """Verify the output of classification tasks."""
        @wraps(func)
        def wrapped_predict(self, dataset):
            response = func(dataset)
            # Must have a prediction for each entry
            assert len(response) == len(dataset)
            # All probabilities should sum to approx. 1
            assert np.all(np.isclose(response.sum(axis=1), np.ones(len(response))))
            return response
        return func


class RegressionExperiment(Experiment):
    """Base class for regression experiments."""

    def __init__(self, *args, **kwargs):
        raise NotImplementedError


class MatchingExperiment(Experiment):
    """Base class for matching experiments."""

    def __init__(self, *args, **kwargs):
        raise NotImplementedError



from sklearn.ensemble import RandomForestClassifier
import pandas as pd

from enso.experiment.grid_search import GridSearch


class RandomForestCV(GridSearch):
    """Implementation of a grid-search optimized RandomForest."""

    def __init__(self, *args, **kwargs):
        """Initialize internal classifier."""
        super().__init__(*args, **kwargs)
        self.base_model = RandomForestClassifier
        self.param_grid = {
            'n_estimators': [10, 50]
        }



"""
From: https://www.figure-eight.com/data-for-everyone/

Contributors looked at a single sentence and rated its emotional content based on Plutchik’s wheel of emotions. 18 emotional choices were presented to contributors for grading.
"""

from enso.download import generic_download


def convert_score_to_category(score):
    if 1 <= score <= 3:
        return "negative"
    elif 4 <= score <= 6:
        return "neutral"
    else:
        return "positive"


if __name__ == "__main__":
    generic_download(
        url="https://www.figure-eight.com/wp-content/uploads/2016/03/us-economic-newspaper.csv",
        text_column="text",
        target_column="positivity",
        target_transformation=convert_score_to_category,
        filename="Economy.csv"
    )



from sklearn.svm import SVC
import pandas as pd

from enso.experiment.grid_search import GridSearch


class SupportVectorMachineCV(GridSearch):
    """Implementation of a grid-search optimized RBF-SVM."""

    def __init__(self, *args, **kwargs):
        """Initialize internal classifier."""
        super().__init__(*args, **kwargs)
        self.base_model = SVC
        self.param_grid = {
            'C': [0.01, 0.1, 1.0, 10., 100.],
            'gamma': [0.1, 1.0, 10.],
            'probability': [True]
        }



import indicoio
import multiprocessing

"""Constants to configure the rest of Enso."""

# Directory for storing data
DATA_DIRECTORY = "Data"

# Directory for storing results
RESULTS_DIRECTORY = "Results"

# Directory for storing features
FEATURES_DIRECTORY = "Features"

# Directory for storing experiment results
EXPERIMENT_NAME = "Demo"

# Datasets to featurize or run experiments on
DATA = {
    'Classify/AirlineSentiment',
    'Classify/MovieReviews',
    'Classify/MPQA',
    'Classify/PoliticalTweetSubjectivity',
}

# Featurizers to activate
FEATURIZERS = {
    "SpacyGloveFeaturizer",
    "SpacyCNNFeaturizer"
}

# Experiments to run
EXPERIMENTS = {
    "LogisticRegressionCV",
}

# Metrics to compute
METRICS = {
    "Accuracy",
    "MacroRocAuc",
}

# Test setup metadata
TEST_SETUP = {
    "train_sizes": range(50, 550, 50),
    "n_splits": 25,
    "samplers": ['Random'],
    "sampling_size": .3,
    "resamplers": ['RandomOverSampler']
}

# Visualizations to display
VISUALIZATIONS = {
    'FacetGridVisualizer'
}

# kwargs to pass directly into visualizations
VISUALIZATION_OPTIONS = {
    'display': True,
    'save': True,
    'FacetGridVisualizer': {
        'x_tile': 'Metric',
        'y_tile': 'Dataset',
        'x_axis': 'TrainSize',
        'y_axis': 'Result',
        'lines': ['Experiment', 'Featurizer'],
        'category': 'merge',
        'cv': 'mean',
        'filename': 'TestResult'
    }
}


N_GPUS = 3
N_CORES = 1 # multiprocessing.cpu_count()



"""
From: https://www.figure-eight.com/data-for-everyone/
A sentiment analysis job about the problems of each major U.S. airline. Twitter data was scraped from February of 2015 and contributors were asked to first classify positive, negative, and neutral tweets, followed by categorizing negative reasons (such as “late flight” or “rude service”).
"""

from enso.download import generic_download


if __name__ == "__main__":
    generic_download(
        url="https://www.figure-eight.com/wp-content/uploads/2016/03/Airline-Sentiment-2-w-AA.csv",
        text_column="text",
        target_column="negativereason",
        filename="AirlineNegativity.csv"
    )
    generic_download(
        url="https://www.figure-eight.com/wp-content/uploads/2016/03/Airline-Sentiment-2-w-AA.csv",
        text_column="text",
        target_column="airline_sentiment",
        filename="AirlineSentiment.csv"
    )



"""
Entrypoint for creating visualizations according to the specifications of `config.py`.
"""
import logging
import abc
import ast
import json
from functools import wraps
import os.path

import pandas as pd
import numpy as np
import matplotlib

from enso.config import VISUALIZATIONS, VISUALIZATION_OPTIONS, RESULTS_DIRECTORY
from enso.utils import get_plugins, get_all_experiment_runs
from enso.utils import BaseObject

# set matplotlib up for headless environment
matplotlib.use("Agg")


class Visualization(object):
    """
    Orchestrates the application of a set of visualizers to a given result file
    using the settings specified in `config.py`
    """

    def __init__(self, test_run=None):
        """Visualize results from a given test run."""
        self.results_id = self._resolve_results_id(test_run)
        self.results = self._grab_results(self.results_id)
        self.visualizers = get_plugins("visualize", VISUALIZATIONS)

    @staticmethod
    def _resolve_results_id(test_run):
        """
        Grab the appropriate results file id.
        `test_run` can be set either to None, which will default to the most recent run,
        a string, which will search for a match timestamp in the results directory, or
        an integer `n` which will grab the test run from `n` runs ago. 0 would be the same as None.
        """
        correct_run = None
        if isinstance(test_run, str):
            correct_run = test_run
            if not os.path.exists(os.path.join(RESULTS_DIRECTORY, correct_run)):
                raise ValueError("Experiment run: %s not found" % test_run)
        elif test_run is None:
            all_runs = get_all_experiment_runs()
            correct_run = all_runs[0]
        elif isinstance(test_run, int):
            correct_run = all_runs[test_run]
        else:
            raise ValueError("test_run must be either None, an int, or a string")
        return correct_run

    @staticmethod
    def _grab_results(results_id):
        """
        Read data file that corresponds with the provided results id and return
        resulting pd.DataFrame() instance.
        """
        df = pd.read_csv('{}/{}/Results.csv'.format(RESULTS_DIRECTORY, results_id))
        try:
            df.drop(columns=["Unnamed: 0"], axis=1, inplace=True)
        except ValueError:
            # no problem
            df['ID'] = df['ID'].astype(str)
            pass
        return df

    def visualize(self):
        """
        Provided with a detailed configuration from `config.py`, iterates over all available :class:`Visualizer`'s and
        runs their respective :func:`Visualizer.visualize` methods using the provided settings.
        """
        global_options = {a: b for a, b in VISUALIZATION_OPTIONS.items() if not isinstance(b, dict)}
        global_options['results'] = self.results
        global_options['results_id'] = self.results_id
        for visualizer in self.visualizers:
            local_options = VISUALIZATION_OPTIONS.get(visualizer.name(), {})
            local_options.update(global_options)
            visualizer.visualize(**local_options)


class Visualizer(BaseObject):
    """
    Base class for creating visualizations.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @abc.abstractmethod
    def visualize(self, results, **kwargs):
        """
        Create visualization for the given test_run.
        Setting `display` to `True` will default to showing the generated visualizations as they are created,
        while setting `write` to `True` will default to saving the generated image in the Results directory.

        :param results: pd.DataFrame of results, loaded from results .csv file.
        """
        raise NotImplementedError


class DataHandler(abc.ABCMeta):
    """
    Handles cross-validation and category strategies on results before passing them on.
    """

    def __new__(cls, *args, **kwargs):
        """Decorate visualize method with cv and category handlers."""
        viz_class = super(DataHandler, cls).__new__(cls, *args, **kwargs)
        viz_class.visualize = viz_class.handle_categories(viz_class.visualize)
        viz_class.visualize = viz_class.handle_cv(viz_class.visualize)
        return viz_class


class ClassificationVisualizer(Visualizer):
    """
    Base class for classification visualizations.
    """

    __metaclass__ = DataHandler

    @classmethod
    def handle_categories(cls, func):
        """
        Execute a category strategy on a result set.
        Force the user to make a choice about handling predictions for different classes.
        It should support either being one of the axes for the Main Visualization,
        or there should be a strategy for turning multiple entries into a single one
        """
        @wraps(func)
        def wrapped_visualize(self, results, **kwargs):
            category = kwargs.get('category', None)
            if category is not None:
                class_info = [ast.literal_eval(result) for result in results['Result']]
                new_info = []
                result = results.copy()
                if category == 'merge':
                    for entry in class_info:
                        sample_value = entry.values()[0]
                        # Check if all entries are basically identical
                        if np.allclose(entry.values(), sample_value):
                            new_info.append(sample_value)
                        else:
                            raise ValueError("`merge` is not a valid strategy if classes vary.")
                    result['Result'] = new_info
                else:
                    raise ValueError("`%s` strategy not found" % category)
                return func(results, **kwargs)
            elif 'Category' not in kwargs.values():
                raise ValueError("""
                    Category must either have a strategy listed, or be a displayed variable.
                """)
        return func

    @classmethod
    def handle_cv(cls, func):
        """
        Execute a cv strategy on a result_set.
        Same as handle_categories, but for multiple cv runs rather than multiple classes
        """
        @wraps(func)
        def wrapped_visualize(self, results, **kwargs):
            cv = kwargs.get('cv', None)
            if cv is not None:
                new_results = pd.DataFrame(columns=results.columns.values)
                for row_set in self._iterate_identical_rows(results):
                    sample_row = row_set.iloc[0]
                    if cv == "mean":
                        if row_set['Result'].dtype == np.float64:
                            mean = row_set['Result'].mean()
                            sample_row['Result'] = mean
                        elif row_set['Result'].dtype == np.object:
                            values = [ast.literal_eval(item) for item in row_set['Result']]
                            average_dict = {}
                            # Grabbing from first entry since they should all be identical
                            for category in values[0]:
                                all_values = [value[category] for value in values]
                                average_dict[category] = sum(all_values) / len(all_values)
                            sample_row['Result'] = json.dumps(average_dict)
                        new_results = new_results.append(sample_row, ignore_index=True)
                    else:
                        raise ValueError("mean is the only cv thing we support right now")
                return func(new_results, **kwargs)
            elif 'Split' not in kwargs.values():
                raise ValueError("""
                    Split must either have a strategy listed, or be a displayed variable.
                """)
        return func

    @classmethod
    def _iterate_identical_rows(self, results):
        """Generate each set of rows identical other than their results column."""
        found_rows = set()
        for _, row in results.iterrows():
            current_row = self._row_repr(row)
            if current_row not in found_rows:
                found_rows.add(current_row)
                yield results.loc[
                    (results['Dataset'] == row['Dataset']) &
                    (results['Experiment'] == row['Experiment']) &
                    (results['Metric'] == row['Metric']) &
                    (results['TrainSize'] == row['TrainSize']) &
                    (results['Featurizer'] == row['Featurizer']) &
                    (results['Sampler'] == row['Sampler']) &
                    (results['Resampling'] == row['Resampling'])
                ]

    @staticmethod
    def _row_repr(row):
        """Return a tuple representation of a row without the Result colum."""
        return (
            row['Dataset'],
            row['Experiment'],
            row['Metric'],
            row['TrainSize'],
            row['Featurizer'],
            row['Sampler'],
            row['Resampling']
        )



"""
From: https://www.figure-eight.com/data-for-everyone/

Contributors looked at over 10,000 tweets culled with a variety of searches like “ablaze”, “quarantine”, and “pandemonium”, then noted whether the tweet referred to a disaster event (as opposed to a joke with the word or a movie review or something non-disastrous).
"""

from enso.download import generic_download

if __name__ == "__main__":
    generic_download(
        url="https://www.figure-eight.com/wp-content/uploads/2016/03/socialmedia-disaster-tweets-DFE.csv",
        text_column="text",
        target_column="choose_one",
        filename="SocialMediaDisasters.csv"
    )



#!/usr/bin/env python
import os
from setuptools import setup, find_packages

REQUIREMENTS = open('requirements.txt').readlines()

setup(
    name='enso',
    version='0.1.4',
    description='Testing feature-sets and learning algorithms for transfer learning.',
    author='indico',
    author_email='engineering@indico.io',
    packages=find_packages(),
    install_requires=REQUIREMENTS
)


from enso.experiment import Experiment
from enso.featurize import Featurizer
from enso.visualize import Visualizer
from enso.sample import Sampler
from enso.metrics import Metric


__version__, version, VERSION = ('0.1.4',) * 3



"""
From: https://www.figure-eight.com/data-for-everyone/

Movie review task from SentEval.  Note that performance on this dataset is not comparable to official SentEval scores because of differences in data splitting.
"""

from enso.download import generic_download

if __name__ == "__main__":
    generic_download(
        url="https://s3.amazonaws.com/enso-data/MovieReviews.csv",
        text_column="Text",
        target_column="Target",
        filename="MovieReviews.csv"
    )



"""Basic classification metrics."""
from sklearn.metrics import roc_auc_score, accuracy_score, log_loss
import numpy as np

from enso.metrics import ClassificationMetric
from enso.utils import labels_to_binary


class WeightedRocAuc(ClassificationMetric):
    """Classwise ROC AUC metric."""

    def evaluate(self, ground_truth, result):
        """Return AUC metric."""
        classwise_auc = {}
        binary_labels = labels_to_binary(ground_truth)
        binary_labels = np.hstack([binary_labels[column].values.reshape(-1, 1) for column in result.columns])
        predicted_labels = np.hstack([result[column].values.reshape(-1, 1) for column in result.columns])
        return roc_auc_score(binary_labels, predicted_labels, average='weighted')


class MacroRocAuc(ClassificationMetric):
    """Classwise ROC AUC metric."""

    def evaluate(self, ground_truth, result):
        """Return AUC metric."""
        classwise_auc = {}
        binary_labels = labels_to_binary(ground_truth)
        binary_labels = np.hstack([binary_labels[column].values.reshape(-1, 1) for column in result.columns])
        predicted_labels = np.hstack([result[column].values.reshape(-1, 1) for column in result.columns])
        return roc_auc_score(binary_labels, predicted_labels, average='macro')


class Accuracy(ClassificationMetric):
    """Return class accuracy"""
    def evaluate(self, ground_truth, result):
        predicted = result.idxmax(axis=1)
        return accuracy_score(ground_truth, predicted)


class LogLoss(ClassificationMetric):
    """LogLoss"""

    def evaluate(self, ground_truth, result):
        """Return AUC metric."""
        binary_labels = labels_to_binary(ground_truth)
        binary_labels = np.hstack([binary_labels[column].values.reshape(-1, 1) for column in result.columns])
        predicted_labels = np.hstack([result[column].values.reshape(-1, 1) for column in result.columns])
        return log_loss(binary_labels, predicted_labels)



"""
From: https://www.figure-eight.com/data-for-everyone/
Contributors looked at thousands of social media messages from US Senators and other American politicians to classify their content. Messages were broken down into audience (national or the tweeter’s constituency), bias (neutral/bipartisan, or biased/partisan), and finally tagged as the actual substance of the message itself (options ranged from informational, announcement of a media appearance, an attack on another candidate, etc.)
"""

from enso.download import generic_download

if __name__ == "__main__":
    generic_download(
        url="https://www.figure-eight.com/wp-content/uploads/2016/03/Political-media-DFE.csv",
        text_column="text",
        target_column="message",
        filename="PoliticalTweetClassification.csv"
    )

    generic_download(
        url="https://www.figure-eight.com/wp-content/uploads/2016/03/Political-media-DFE.csv",
        text_column="text",
        target_column="bias",
        filename="PoliticalTweetBias.csv"
    )



"""
From: https://www.figure-eight.com/data-for-everyone/
A Twitter topic analysis of users’ 2015 New Year’s resolutions.
"""

from enso.download import generic_download

if __name__ == "__main__":
    generic_download(
        url="https://www.figure-eight.com/wp-content/uploads/2016/03/New-years-resolutions-DFE.csv",
        text_column="text",
        target_column="Resolution_Category",
        filename="NewYearsResolutions.csv"
    )



"""
Download all available datasets
"""
import glob
import importlib
import os.path
import runpy
import logging

if __name__ == "__main__":
    glob_path = os.path.join(os.path.dirname(__file__), "*.py")
    filenames = [
        os.path.basename(f) for f in glob.glob(glob_path)
        if not os.path.basename(f).startswith("_")
    ]
    for filename in filenames:
        module_name = filename.split('.')[0]
        print("\n$ python -m enso.download.{}".format(module_name))
        runpy.run_module("enso.download.{}".format(module_name), run_name='__main__')



from sklearn.naive_bayes import GaussianNB, MultinomialNB
import pandas as pd

from enso.experiment import Experiment


class NaiveBayes(Experiment):
    """Gaussian naive bayes model."""

    param_grid = {}

    def __init__(self, *args, **kwargs):
        """Initialize internal classifier."""
        super().__init__(*args, **kwargs)
        self.model = GaussianNB()

    def fit(self, X, y):
        self.model.fit(X, y)

    def predict(self, X, **kwargs):
        """Predict results on test set based on current internal model."""
        labels = self.model.classes_
        probabilities = self.model.predict_proba(X)
        return pd.DataFrame({label: probabilities[:, i] for i, label in enumerate(labels)})



from . import Sampler
import numpy as np


class Orthogonal(Sampler):
    """
    Randomly selects starting points, then selects additional points for which
    the product of the cosine distance to all starting points is maximally large.
    Each selected point is then iteratively added to the list of starting points.

    :param data: pd.Series of feature vectors
    :param train_labels: pd.Series of targets
    :param train_indices: pd.Series of example indices
    :param train_size: int number of examples to select
    """
    DISTANCE_FUNCTION = "cosine"

    def sample(self):
        """
        Applies the orthogonal sampling strategy.

        :returns: np.array of example indices selected by `Orthogonal`.
        """
        centers = self._choose_starting_points()
        while len(centers) < self.train_size:
            reduced = np.prod(self.distances[centers], 0)
            center = np.argmax(np.absolute(reduced))
            centers.append(center)
        return centers


"""
From: https://www.figure-eight.com/data-for-everyone/
Contributors evaluated tweets about multiple brands and products. The crowd was asked if the tweet expressed positive, negative, or no emotion towards a brand and/or product. If some emotion was expressed they were also asked to say which brand or product was the target of that emotion.
"""

from enso.download import generic_download


if __name__ == "__main__":
    generic_download(
        url="https://www.figure-eight.com/wp-content/uploads/2016/03/judge-1377884607_tweet_product_company.csv",
        text_column="tweet_text",
        target_column="is_there_an_emotion_directed_at_a_brand_or_product",
        filename="BrandEmotion.csv"
    )
    generic_download(
        url="https://www.figure-eight.com/wp-content/uploads/2016/03/judge-1377884607_tweet_product_company.csv",
        text_column="tweet_text",
        target_column="emotion_in_tweet_is_directed_at",
        filename="BrandEmotionCause.csv"
    )



"""
From: https://www.figure-eight.com/data-for-everyone/

Contributors read color-coded sentences and determined what the relationship of a drug was to certain symptoms or diseases. There are two types of relationships. A drug either:

Caused side effects – [Drug] gave me [symptom]
Was effective against a condition – [Drug] helped my [disease]
Is prescribed for a certain disease – [Drug] was given to help my [disease]
Is contraindicated in – [Drug] should not be taken if you have [disease or symptom]
The second similarity was more about the statement itself. Those broke down into:

Personal experiences – I started [drug] for [disease]
Personal experiences negated – [Drug] did not cause [symptom]
Impersonal experiences – I’ve heard [drug] causes [symptom]
Impersonal experiences negated – I’ve read [drug] doesn’t cause [symptom]
Question – Have you tried [drug]?
"""
from enso.download import generic_download, html_to_text


if __name__ == "__main__":
    generic_download(
        url="https://www.figure-eight.com/wp-content/uploads/2016/03/drug-relation-dfe.csv",
        text_column="text",
        target_column="human_relation",
        text_transformation=html_to_text,
        filename="DrugReviewType.csv"
    )

    generic_download(
        url="https://www.figure-eight.com/wp-content/uploads/2016/03/drug-relation-dfe.csv",
        text_column="text",
        target_column="human_relation_type",
        text_transformation=html_to_text,
        filename="DrugReviewIntent.csv"
    )



"""
From: https://www.figure-eight.com/data-for-everyone/

Stanford Sentiment Treebank binary task from SentEval.  Note that performance on this dataset is not comparable to official SentEval scores because of differences in data splitting.
"""

from enso.download import generic_download

if __name__ == "__main__":
    generic_download(
        url="https://s3.amazonaws.com/enso-data/SST-binary.csv",
        text_column="Text",
        target_column="Target",
        filename="SST-binary.csv"
    )



import os.path
import requests
import pandas as pd
from pandas.compat import StringIO
from bs4 import BeautifulSoup

from enso import config


def generic_download(url, text_column, target_column, filename, save=True, task_type='Classify', text_transformation=None, target_transformation=None):

    save_path = os.path.join(config.DATA_DIRECTORY, task_type, filename)
    if os.path.exists(save_path):
        print("{} already downloaded, skipping...".format(filename))
        return

    response = requests.get(url)
    _file = StringIO(response.text.replace('\r', '\n'))
    df = pd.read_csv(_file)
    df = df.dropna(subset=[text_column, target_column])

    new_df = pd.DataFrame(columns=['Text', 'Target'])
    new_df['Text'], new_df['Target'] = df[text_column], df[target_column]

    if text_transformation is not None:
        new_df['Text'] = new_df['Text'].apply(text_transformation)
    if target_transformation is not None:
        new_df['Target'] = new_df['Target'].apply(target_transformation)

    if save:
        new_df.to_csv(save_path, index=False)

    return new_df


def html_to_text(text):
    return BeautifulSoup(text, "html5lib").get_text()



"""
From: https://www.figure-eight.com/data-for-everyone/
A data categorization job concerning what corporations actually talk about on social media. Contributors were asked to classify statements as information (objective statements about the company or it’s activities), dialog (replies to users, etc.), or action (messages that ask for votes or ask users to click on links, etc.).
"""

from enso.download import generic_download

if __name__ == "__main__":
    generic_download(
        url="https://www.figure-eight.com/wp-content/uploads/2016/03/Corporate-messaging-DFE.csv",
        text_column="text",
        target_column="category",
        filename="CorporateMessaging.csv"
    )
